@@ -1,16 +1,39 @@
 use crate::comment_parsing::parse_comments;
 use crate::comment_parsing::CommentContent;
-use pgn_reader::{BufferedReader, RawComment, SanPlus, Skip, Visitor};
+use pgn_reader::{BufferedReader, RawComment, RawHeader, SanPlus, Skip, Visitor};
 use pyo3::prelude::*;
 use rayon::prelude::*;
+use rayon::ThreadPoolBuilder;
+use shakmaty::Color;
 use shakmaty::{uci::UciMove, Chess, Outcome, Position};
 use std::io::Cursor;
 
 mod comment_parsing;
 
+#[pyclass]
+#[derive(Clone)]
+pub struct PositionStatus {
+    #[pyo3(get)]
+    is_checkmate: bool,
+
+    #[pyo3(get)]
+    is_stalemate: bool,
+
+    #[pyo3(get)]
+    legal_move_count: usize,
+
+    #[pyo3(get)]
+    is_game_over: bool,
+
+    #[pyo3(get)]
+    insufficient_material: (bool, bool),
+
+    #[pyo3(get)]
+    turn: bool,
+}
 #[pyclass]
 /// A Visitor to extract SAN moves and comments from PGN movetext
-struct MoveExtractor {
+pub struct MoveExtractor {
     #[pyo3(get)]
     moves: Vec<String>,
 
@@ -29,6 +52,15 @@ struct MoveExtractor {
     #[pyo3(get)]
     outcome: Option<String>,
 
+    #[pyo3(get)]
+    headers: Vec<(String, String)>,
+
+    #[pyo3(get)]
+    castling_rights: Vec<(bool, bool, bool, bool)>,
+
+    #[pyo3(get)]
+    position_status: Option<PositionStatus>,
+
     pos: Chess,
 }
 
@@ -44,13 +76,47 @@ impl MoveExtractor {
             evals: Vec::with_capacity(100),
             clock_times: Vec::with_capacity(100),
             outcome: None,
+            headers: Vec::with_capacity(10),
+            castling_rights: Vec::with_capacity(100),
+            position_status: None,
         }
     }
+
+    fn turn(&self) -> bool {
+        match self.pos.turn() {
+            Color::White => true,
+            Color::Black => false,
+        }
+    }
+
+    fn update_position_status(&mut self) {
+        // TODO this checks legal_moves() a bunch of times
+        self.position_status = Some(PositionStatus {
+            is_checkmate: self.pos.is_checkmate(),
+            is_stalemate: self.pos.is_stalemate(),
+            legal_move_count: self.pos.legal_moves().len(),
+            is_game_over: self.pos.is_game_over(),
+            insufficient_material: (
+                self.pos.has_insufficient_material(Color::White),
+                self.pos.has_insufficient_material(Color::Black),
+            ),
+            turn: match self.pos.turn() {
+                Color::White => true,
+                Color::Black => false,
+            },
+        });
+    }
 }
 
 impl Visitor for MoveExtractor {
     type Result = bool;
 
+    fn header(&mut self, key: &[u8], value: RawHeader<'_>) {
+        let key_str = String::from_utf8_lossy(key).into_owned();
+        let value_str = String::from_utf8_lossy(value.as_bytes()).into_owned();
+        self.headers.push((key_str, value_str));
+    }
+
     fn begin_game(&mut self) {
         self.moves.clear();
         self.pos = Chess::default();
@@ -67,6 +133,17 @@ impl Visitor for MoveExtractor {
                     self.pos.play_unchecked(&m);
                     let uci = UciMove::from_standard(&m);
                     self.moves.push(uci.to_string());
+                    // self.pos.castles().castling_rights();
+                    // println!("castling {:?}", self.pos.castles());
+                    let castling_bitboard = self.pos.castles().castling_rights();
+                    let castling_rights = (
+                        castling_bitboard.contains(shakmaty::Square::A1),
+                        castling_bitboard.contains(shakmaty::Square::H1),
+                        castling_bitboard.contains(shakmaty::Square::A8),
+                        castling_bitboard.contains(shakmaty::Square::H8),
+                    );
+
+                    self.castling_rights.push(castling_rights);
                 }
                 Err(err) => {
                     eprintln!("error in game: {} {}", err, san_plus);
@@ -91,7 +168,7 @@ impl Visitor for MoveExtractor {
                 for content in parsed_comments {
                     match content {
                         CommentContent::Text(text) => {
-                            if text.trim() != "" {
+                            if !text.trim().is_empty() {
                                 move_comments.push_str(&text);
                             }
                         }
@@ -116,7 +193,7 @@ impl Visitor for MoveExtractor {
                         }
                     }
                 }
-                self.comments.push(move_comments); // Add the concatenated comment string
+                self.comments.push(move_comments);
             }
             Err(e) => {
                 eprintln!("Error parsing comment: {:?}", e);
@@ -133,6 +210,7 @@ impl Visitor for MoveExtractor {
             Outcome::Decisive { winner } => format!("{:?}", winner),
             Outcome::Draw => "Draw".to_string(),
         });
+        self.update_position_status();
     }
 
     fn end_game(&mut self) -> Self::Result {
@@ -140,44 +218,61 @@ impl Visitor for MoveExtractor {
     }
 }
 
-fn parse_single_game(pgn: &str) -> PyResult<MoveExtractor> {
+// --- Native Rust versions (no PyResult) ---
+pub fn parse_single_game_native(pgn: &str) -> Result<MoveExtractor, String> {
     let mut reader = BufferedReader::new(Cursor::new(pgn));
     let mut extractor = MoveExtractor::new();
-
     match reader.read_game(&mut extractor) {
         Ok(Some(_)) => Ok(extractor),
-        Ok(None) => Err(pyo3::exceptions::PyValueError::new_err(
-            "No game found in PGN",
-        )),
-        Err(err) => Err(pyo3::exceptions::PyValueError::new_err(format!(
-            "Parsing error: {}",
-            err
-        ))),
+        Ok(None) => Err("No game found in PGN".to_string()),
+        Err(err) => Err(format!("Parsing error: {}", err)),
     }
 }
 
-fn parse_multiple_games(pgns: Vec<String>) -> PyResult<Vec<MoveExtractor>> {
-    let results: Vec<PyResult<MoveExtractor>> =
-        pgns.par_iter().map(|pgn| parse_single_game(pgn)).collect();
+pub fn parse_multiple_games_native(
+    pgns: &Vec<String>,
+    num_threads: Option<usize>,
+) -> Result<Vec<MoveExtractor>, String> {
+    let num_threads = num_threads.unwrap_or_else(|| num_cpus::get());
 
-    // Convert results to a single PyResult, aggregating errors
-    let extractors: Vec<MoveExtractor> = results.into_iter().collect::<Result<Vec<_>, _>>()?;
-    Ok(extractors)
+    // Build a custom Rayon thread pool with the desired number of threads
+    let thread_pool = ThreadPoolBuilder::new()
+        .num_threads(num_threads)
+        .build()
+        .expect("Failed to build Rayon thread pool");
+
+    thread_pool.install(|| {
+        let results: Vec<Result<MoveExtractor, String>> = pgns
+            .par_iter()
+            .map(|pgn| parse_single_game_native(pgn))
+            .collect();
+
+        let mut extractors: Vec<MoveExtractor> = Vec::with_capacity(results.len());
+        for res in results {
+            match res {
+                Ok(extractor) => extractors.push(extractor),
+                Err(e) => return Err(e),
+            }
+        }
+        Ok(extractors)
+    })
 }
 
-/// Parses a full PGN game and returns a list of SAN moves and parsed comments
+// --- Python-facing wrappers (PyResult) ---
 #[pyfunction]
 fn parse_game(pgn: &str) -> PyResult<MoveExtractor> {
-    parse_single_game(pgn)
+    parse_single_game_native(pgn).map_err(|err| pyo3::exceptions::PyValueError::new_err(err))
 }
 
 /// In parallel, parse a set of games
 #[pyfunction]
-fn parse_games(pgns: Vec<String>) -> PyResult<Vec<MoveExtractor>> {
-    parse_multiple_games(pgns)
+#[pyo3(signature = (pgns, num_threads=None))]
+fn parse_games(pgns: Vec<String>, num_threads: Option<usize>) -> PyResult<Vec<MoveExtractor>> {
+    parse_multiple_games_native(&pgns, num_threads)
+        .map_err(|err| pyo3::exceptions::PyValueError::new_err(err))
 }
 
-/// A Python module implemented in Rust.
+// Parser for chess PGN notation
 #[pymodule]
 fn rust_pgn_reader_python_binding(m: &Bound<'_, PyModule>) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(parse_game, m)?)?;
